/* Phenoelit VIPPR - the Virtual IP Phalanx Router 
 *
 * $Id: vippr.c,v 1.5 2001/07/10 04:24:39 fx Exp fx $
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <signal.h>

#include <protocols.h>
#include <build.h>

#include <vippr.h>
#include "vippr_grammar.tab.h"


extern FILE	*yyin;
vip_t		*vip_anchor=NULL;
routers_t	*routers_anchor=NULL;
interface_t	*if_anchor=NULL;
cfg_t		cfg;
sys_t		sys;
arpcache_t	*arp_anchor;

u_char		bcastaddr[6];
struct in_addr	bcastip;


void usage(char *n);
void yyparse(void);
void signalhndlr(int s);


int main(int argc, char **argv) {
    interface_t	*intf;
    char	option;
    extern char	*optarg;

    if ((unsigned int)geteuid()!=0) { 
	fprintf(stderr,"Become r00t\n"); 
	return 1;
    }
    memset(&cfg,0,sizeof(cfg));
    memset(&sys,0,sizeof(sys));

    /* read command line */
    while ((option=getopt(argc,argv,"viFDf:"))!=EOF) {
	switch (option) {
	    case 'v':	cfg.verbose++;
			break;
	    case 'D':	cfg.debug++;
			break;
	    case 'f':	cfg.cfgfile=smalloc(strlen(optarg)+1);
			strcpy(cfg.cfgfile,optarg);
			break;

	    case 'i':	cfg.disable_interface_add++;
			break;
	    case 'F':	cfg.disable_fastdrop++;
			break;
	    default:	usage(argv[0]);
			break;
	} // end of switch
    } // end of while


    /* check defaults */
    if (cfg.cfgfile==NULL) { 
	cfg.cfgfile=smalloc(strlen(VIPPR_DEFAULT_CFGFILE)+1);
	strcpy(cfg.cfgfile,VIPPR_DEFAULT_CFGFILE);
    }
    
    
    printf("Phenoelit VIPPR [Virtual IP Phalanx Router] Build %s\n"
	    "by FX <fx@phenoelit.de> of Phenoelit (http://www.phenoelit.de/)\n",
	    BUILD);


    /* read config */
    if ((yyin=fopen(cfg.cfgfile,"rt"))==NULL) {
	fprintf(stderr,"Unable to open config file %s\n",cfg.cfgfile);
	return (-1);
    } else {
	yyparse();
    }

    /* --- at this point, ip addresses and routing tables are in the system 
     *     and the interfaces are up and running */

    /* get mem for the packet buffer and set stdaddrs*/
    sys.packet=smalloc(PACKET_MAX);
    memset(bcastaddr,0xFF,6);
    memset(&(bcastip.s_addr),0xFF,IP_ADDR_LEN);

    if ((intf=if_anchor)==NULL) {
	fprintf(stderr,"main(): Ooops.. no interfaces ?\n");
	return 1;
    }

    /* set up the memory shortcut links between VIP structres and Routers */
    if (router_linktovip()!=0) return 1;
    if (cfg.verbose) router_print_tables();

    signal(SIGINT,&signalhndlr);
    signal(SIGTERM,&signalhndlr);
    /* *****************************
     *    MAIN L00P 
     * *****************************/

    if ((cfg.debug)||(cfg.verbose>2)) printf("main(): entering main mode\n");
    while (!sys.ctrlc) {
	/* one in, one out */
	net_ifrecv(intf);
	net_ifsend(intf);

	/* loop through the interfaces */
	if (intf->next!=NULL) intf=intf->next; 
	else {
	    intf=if_anchor;
	    check_arp_table();
	}

	// usleep(1);
    }
    print_arp_table();

    /* *****************************
     *    MAIN L00P END
     * *****************************/

    vip_destroy();
    interface_destroy();
    return 0;
}




/*****************************************************************
 *
 *  ADDITIONAL STUFF
 *
 *****************************************************************/

void *smalloc(size_t s) {
    void	*p;

    p=malloc(s);
    if (p==NULL) {
	usleep(1000);
	p=malloc(s);
	if (p==NULL) {
	    fprintf(stderr,"smalloc(): FATAL - memory allocation failed\n");
	    exit(-127);
	}
    }
    memset(p,0,s);
    return p;
}

void yyerror(char *s) {
    extern int yylineno;
    extern char yytext[];

    fprintf(stderr,"VIPPR config parser: %s in line %d "
	    "while parsing text '%s'\n",s,yylineno,yytext);
    exit (-1);
}

void signalhndlr(int s) {
    sys.ctrlc++;
}

/* most boring stuff in the world */
void usage(char *n) {
    fprintf(stderr,"%s [-v[v[v]]] [-D] [-f <config.file>] [-i] [-F]\n"
	    "\t-i disables automatic adding of VIPs to their router process\n"
	    "\t-D enables debug output\n",n);
    exit(-1);
}

/* A better version of hdump, from Lamont Granquist.  Modified slightly
   by Fyodor (fyodor@DHP.com). Obviously stolen from nmap (util.c) by FX */
void lamont_hdump(unsigned char *bp, unsigned int length) {

  /* stolen from tcpdump, then kludged extensively */

  static const char asciify[] = "................................ !\"#$%&'()*+,-./0123456789:;<=>?@ABCDEFGHIJKLMNOPQRSTUVWXYZ[\\]^_`abcdefghijklmnopqrstuvwxyz{|}~.................................................................................................................................";

  register const u_short *sp;
  register const u_char *ap;
  register u_int i, j;
  register int nshorts, nshorts2;
  register int padding;

  printf("\n\t");
  padding = 0;
  sp = (u_short *)bp;
  ap = (u_char *)bp;
  nshorts = (u_int) length / sizeof(u_short);
  nshorts2 = (u_int) length / sizeof(u_short);
  i = 0;
  j = 0;
  while(1) {
    while (--nshorts >= 0) {
      printf(" %04x", ntohs(*sp));
      sp++;
      if ((++i % 8) == 0)
        break;
    }
    if (nshorts < 0) {
      if ((length & 1) && (((i-1) % 8) != 0)) {
        printf(" %02x  ", *(u_char *)sp);
        padding++;
      }
      nshorts = (8 - (nshorts2 - nshorts));
      while(--nshorts >= 0) {
        printf("     ");
      }
      if (!padding) printf("     ");
    }
    printf("  ");

    while (--nshorts2 >= 0) {
      printf("%c%c", asciify[*ap], asciify[*(ap+1)]);
      ap += 2;
      if ((++j % 8) == 0) {
        printf("\n\t");
        break;
      }
    }
    if (nshorts2 < 0) {
      if ((length & 1) && (((j-1) % 8) != 0)) {
        printf("%c", asciify[*ap]);
      }
      break;
    }
  }
  if ((length & 1) && (((i-1) % 8) == 0)) {
    printf(" %02x", *(u_char *)sp);
    printf("                                       %c", asciify[*ap]);
  }
  printf("\n");
}
