/* VIPPR
 * Configuration Lexer
 *
 * $Id: vippr_format.y,v 1.3 2001/06/12 14:25:51 fx Exp fx $
 */

/* C-Style implementations */
%{
#include <stdlib.h>
#include <string.h>
#include "vippr.h"
#include "vippr_grammar.tab.h"

/* private stuff - for the string's hack
 * SORRY for that, but sometimes, lexer seem to be very strange */
 int     i;
 char    c,c2,tstr[STRING_FIX_LENGTH];
%}

DIGIT	[0-9]
IDENT	[a-zA-Z][a-zA-Z0-9_]+
PINT	[0-9]+
IPADDR	[0-9]{1,3}\.[0-9]{1,3}\.[0-9]{1,3}\.[0-9]{1,3}
MACADDR	[0-9A-F]{12}

%array
%option yylineno

%x comment
%x thestring

%%	/* RULES SECTION */

	/* keywords */
"VIP"		{ return VIP; }
"ROUTER"	{ return ROUTER; }
"stealth"	{ return VIP_STEALTH; }
"gre"		{ return VIP_GRE; }
"802.1q"	{ return VIP_VLAN8021Q; }

	/* not symbols - real seperators */
"{"{1}	return '{';
"}"{1}	return '}';
"<"{1}	return '<';
">"{1}	return '>';
";"{1}	return ';';
","{1}	return ',';
"="{1}	return '=';
"/"{1}	return '/';

        /* mac address */
{MACADDR}       {
                memset(&yylval.mac_addr,0,13);
                strncpy(yylval.mac_addr,yytext,12);
                return MACADDR;
                }
        /* ip address */
{IPADDR}        {
                memset(&yylval.ip_addr,0,16);
                strncpy(yylval.ip_addr,yytext,15);
                return IPADDR;
                }
        /* integer value */
{PINT}          {
                yylval.i_val = atoi(yytext);
                return PINT;
                }

        /* IDENT string */
{IDENT}         {
                memset(&yylval.p_str,0,STRING_FIX_LENGTH);
                strncpy(yylval.p_str,yytext,STRING_FIX_LENGTH-1);
                return IDENT;
                }


	/* all spaces fall in the hole ;) */
[ \t\n]+	/* empty */


	/* kill all comments */

"/*"	BEGIN(comment);
<comment>[^*]*		/* kill all except **/
<comment>"*"+[^*/]*	/* kill *'s not followed by / or * */
<comment>"*/"	BEGIN(INITIAL);


        /* Strings enclosed in " " */
"\""    {
                memset(&tstr,0,STRING_FIX_LENGTH);
                c=c2='\0';
                i=0;
        UglyAgain:
                while (((c=input())!='"')
                        &&(i<STRING_FIX_LENGTH-2)) {
                        if (c==EOF) {
                            printf("Syntax error on line #%d\n"
                                "\tUnexpeced end of file (open string ?)\n",
                                yylineno);
                            yyterminate();
                        }
                        tstr[i++]=c;
                        c2=c;
                }
                if (i>=STRING_FIX_LENGTH-2) {
                        c2='\0';        /* prevent the ESC-Hack from working*/
                        while (input()!='"');
                }
                /* was it escaped ?? */
                if ((c2=='\\')
                        &&(i<STRING_FIX_LENGTH-2)) {
                        tstr[i-1]=c;
                        goto UglyAgain;
                }

                memset(&yylval.p_str,0,STRING_FIX_LENGTH);
                strncpy(yylval.p_str,tstr,STRING_FIX_LENGTH-1);
                return ASTRING;
        }


	/*fallback character */
. 	{ 
	printf("Syntax error on line #%d\n'%s' is unknown\n",yylineno,yytext); 
	yyterminate();
	}

%%
