/* VIPPR vip management functions 
 *
 * $Id: vippr_mgmt.c,v 1.3 2001/06/12 14:25:51 fx Exp $
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <time.h>

#include <protocols.h>
#include <vippr.h>

/* ****************
 * VIPs 
 * ****************/
int vip_add_stealth(char *ipaddr, char *netmask, char *mac, 
	unsigned int router, char *devicename) {
    vip_t	*new,*last;
    interface_t	*intf;

    /* first make sure the interface exisits and is in the list */
    if ((intf=interface_add(devicename))==NULL) {
	yyerror("vip_add_stealth(): Interface invalid"); 
    }

    /* have a new structure */
    new=smalloc(sizeof(vip_t));
   
    if (inet_aton(ipaddr,&(new->addr))==0) { 
	yyerror("vip_add_stealth(): IPaddress invalid"); 
    }
    if (inet_aton(netmask,&(new->mask))==0) { 
	yyerror("vip_add_stealth(): IPaddress invalid"); 
    }
    sscanf(mac,"%02x%02x%02x%02x%02x%02x", 
	    (unsigned int *)&(new->mac.ether_addr_octet[0]), 
	    (unsigned int *)&(new->mac.ether_addr_octet[1]), 
	    (unsigned int *)&(new->mac.ether_addr_octet[2]), 
	    (unsigned int *)&(new->mac.ether_addr_octet[3]), 
	    (unsigned int *)&(new->mac.ether_addr_octet[4]), 
	    (unsigned int *)&(new->mac.ether_addr_octet[5]));
    new->router=router; 
    new->intf=intf;

    if (vip_anchor==NULL) { 
	vip_anchor=new; 
    } else {
	last=vip_anchor;
	while (last->next!=NULL) last=last->next;
	last->next=new;
    }
    
    if ((cfg.debug)||(cfg.verbose>2)) {
	printf("vip_add_stealth(): \n\tVIP %s",inet_ntoa(new->addr));
	printf("/%s at %x:%x:%x:%x:%x:%x (Rtr %d, %s)\n",
	    inet_ntoa(new->mask),
	    new->mac.ether_addr_octet[0],
	    new->mac.ether_addr_octet[1],
	    new->mac.ether_addr_octet[2],
	    new->mac.ether_addr_octet[3],
	    new->mac.ether_addr_octet[4],
	    new->mac.ether_addr_octet[5],
	    new->router,new->intf->name);
    }

    return 0;
}

int vip_add_gre(char *ipaddr, char *netmask, char *mac, 
	unsigned int router, char *devicename, 
	char *tunnelsrc, char *tunneldest, int usekey, u_int32_t key) {
    vip_t	*new,*last;
    interface_t	*intf;

    /* first make sure the interface exisits and is in the list */
    if ((intf=interface_add(devicename))==NULL) {
	yyerror("vip_add_gre(): Interface invalid"); 
    }

    /* have a new structure */
    new=smalloc(sizeof(vip_t));
   
    if (inet_aton(ipaddr,&(new->addr))==0) { 
	yyerror("vip_add_gre(): IPaddress invalid"); 
    }
    if (inet_aton(netmask,&(new->mask))==0) { 
	yyerror("vip_add_gre(): IPaddress invalid"); 
    }
    sscanf(mac,"%02x%02x%02x%02x%02x%02x", 
	    (unsigned int *)&(new->mac.ether_addr_octet[0]), 
	    (unsigned int *)&(new->mac.ether_addr_octet[1]), 
	    (unsigned int *)&(new->mac.ether_addr_octet[2]), 
	    (unsigned int *)&(new->mac.ether_addr_octet[3]), 
	    (unsigned int *)&(new->mac.ether_addr_octet[4]), 
	    (unsigned int *)&(new->mac.ether_addr_octet[5]));
    new->router=router; 
    new->intf=intf;
    /* GRE special settings */
    new->viptype=VIP_TYPE_GRE;
    if (inet_aton(tunnelsrc,&(new->ex.gre.tunnelsource))==0) { 
	yyerror("vip_add_gre(): tunnel source invalid"); 
    }
    if (inet_aton(tunneldest,&(new->ex.gre.tunneldestination))==0) { 
	yyerror("vip_add_gre(): tunnel destination invalid"); 
    }
    if (usekey) {
	new->ex.gre.flags=VIP_GRE_TUNNELKEY;
	new->ex.gre.tunnelkey=key;
    }

    if (vip_anchor==NULL) { 
	vip_anchor=new; 
    } else {
	last=vip_anchor;
	while (last->next!=NULL) last=last->next;
	last->next=new;
    }
    
    if ((cfg.debug)||(cfg.verbose>2)) {
	printf("vip_add_gre(): \n\tVIP %s",inet_ntoa(new->addr));
	printf("/%s at %x:%x:%x:%x:%x:%x (Rtr %d, %s)\n",
	    inet_ntoa(new->mask),
	    new->mac.ether_addr_octet[0],
	    new->mac.ether_addr_octet[1],
	    new->mac.ether_addr_octet[2],
	    new->mac.ether_addr_octet[3],
	    new->mac.ether_addr_octet[4],
	    new->mac.ether_addr_octet[5],
	    new->router,new->intf->name);
	printf("\tTunnel intrustion: GRE, %s",
		inet_ntoa(new->ex.gre.tunnelsource));
	printf(" >==> %s\n",
		inet_ntoa(new->ex.gre.tunneldestination));
	if (usekey) printf("\tTunnelkey: %u\n",new->ex.gre.tunnelkey);
    }

    return 0;
}

vip_t *vip_findbyip(struct in_addr *ip) {
    vip_t	*c;

    c=vip_anchor;
    while (c!=NULL) 
	if (!memcmp(&(c->addr),ip,IP_ADDR_LEN)) return c; else c=c->next;
    return NULL;
}

vip_t *vip_findbyipif(struct in_addr *ip,interface_t *intf) {
    vip_t	*c;

    c=vip_anchor;
    while (c!=NULL) 
	if ((!memcmp(&(c->addr),ip,IP_ADDR_LEN))&&(c->intf==intf)) 
	    return c; else c=c->next;
    return NULL;
}

vip_t *vip_findbymac(struct ether_addr *mac) {
    vip_t	*c;

    c=vip_anchor;
    while (c!=NULL) 
	if (!memcmp(&(c->mac),mac,ETH_ALEN)) return c; else c=c->next;
    return NULL;
}

vip_t *vip_findbymacif(struct ether_addr *mac,interface_t *intf) {
    vip_t	*c;

    c=vip_anchor;
    while (c!=NULL) 
	if ((!memcmp(&(c->mac),mac,ETH_ALEN))&&(c->intf==intf))
	    return c; else c=c->next;
    return NULL;
}

void vip_destroy(void) {
    vip_t	*c;

    c=vip_anchor;
    while (c!=NULL) {
	vip_anchor=c;
	c=c->next;
	free(vip_anchor);
    }
}


/* *********************
 * ROUTERS 
 * *********************/

routers_t *router_findbynum(unsigned int router) {
    routers_t	*r;

    r=routers_anchor;
    while (r!=NULL) 
	if (r->id==router) return r; else r=r->next;
    return NULL;
}

int router_add(unsigned int rid) {
    routers_t	*r,*c;

    if (router_findbynum(rid)!=NULL) return (-1);
    r=smalloc(sizeof(routers_t));
    if (routers_anchor==NULL) {
	routers_anchor=r;
    } else {
	c=routers_anchor;
	while (c->next!=NULL) c=c->next;
	c->next=r;
    }
    r->id=rid;

    if ((cfg.debug)||(cfg.verbose>2)) 
	printf("router_add(): added router ID %d\n",r->id);

    return 0;
}
    
int router_add_route(unsigned int rid, char *ip, char *mask, char *gw) {
    route_t	*rt,*ct;
    routers_t	*r;

    if ((r=router_findbynum(rid))==NULL) {
	fprintf(stderr,"router_add_route(): failed - %d no such router\n",rid);
	return (-1);
    }

    rt=smalloc(sizeof(route_t));

    if (inet_aton(ip,(struct in_addr *)&(rt->dest))==0) { 
	yyerror("router_add_route(): IPaddress invalid"); 
    }
    if (inet_aton(mask,(struct in_addr *)&(rt->mask))==0) {
	yyerror("router_add_route(): IPmask invalid"); 
    }
    /* make it the beginning of the net */
    rt->dest=rt->dest&rt->mask;
    if (inet_aton(gw,&(rt->gwip))==0) {
	yyerror("router_add_route(): IPgw invalid");
    }
    
    if (r->rtr==NULL) {
	r->rtr=rt;
    } else {
	ct=r->rtr;
	while (ct->next!=NULL) ct=ct->next;
	ct->next=rt;
    }

    if ((cfg.debug)||(cfg.verbose>2)) {
	printf("router_add_route(): [%2d] added %s",r->id,
		inet_ntoa(*((struct in_addr *)&(rt->dest))));
	printf("/%s",inet_ntoa(*((struct in_addr *)&(rt->mask))));
	printf(" -> %s\n",inet_ntoa(rt->gwip));
    }

    arp_new(&(rt->gwip));

    return 0;
}

/* this function basicly goes through all router entries and for each one takes
 * all VIPs and tries to fill the rtab component so that vip structures can
 * refer to their routers as vipstruct->rtab. 
 * If there is a VIP with rid!=0 but rtab==NULL, the router is missing and the
 * function returns -1, 0 otherwise.
 */
int router_linktovip(void) { 
    routers_t	*r;
    route_t	*rr;
    vip_t	*v;
    int		i;
    int		rc=0;

    if (!cfg.disable_interface_add) {
	/* insert interface-network/netmask and 
	 * interface IP into routing tables */
	r=routers_anchor;
	while (r!=NULL) {
	    v=vip_anchor;
	    while (v!=NULL) {
		if (r->id==v->router) {
		    rr=smalloc(sizeof(route_t));
		    memcpy((struct in_addr *)&(rr->dest),
			    &(v->addr),IP_ADDR_LEN);
		    memcpy((struct in_addr *)&(rr->mask),
			    &(v->mask),IP_ADDR_LEN);
		    memcpy((struct in_addr *)&(rr->gwip),
			    &(v->addr),IP_ADDR_LEN);
		    rr->dest=rr->dest&rr->mask;
		    /* route entry points to next route entry */
		    rr->next=r->rtr;
		    /* router entry points to new route */
		    r->rtr=rr;
		}
		v=v->next;
	    }
	    r=r->next;
	}
    }

    /* link routers into vip structures */
    r=routers_anchor;
    while (r!=NULL) {
	i=0;
	v=vip_anchor;
	while (v!=NULL) {
	    if (r->id==v->router) {
		if (cfg.verbose>2) 
		    printf("router_linktovip(): linking VIP %s to Router %d\n",
			    inet_ntoa(v->addr),r->id);
		v->rtab=r->rtr;
		i++;
	    }
	    v=v->next;
	}
	if (i==0) fprintf(stderr,"WARNING: Router %d is unused\n",r->id);
	r=r->next;
    }

    /* check for loose/not-existing routers */
    v=vip_anchor;
    while (v!=NULL) {
	if ((v->router!=0)&&(v->rtab==NULL)) {
	    fprintf(stderr,"ERROR: VIP %s uses undefined router %d\n",
		    inet_ntoa(v->addr),v->router);
	    rc=(-1);
	}
	v=v->next;
    }

    return (rc);
}

void router_print_tables(void) {
    routers_t	*r;
    route_t	*rr;

    printf("--- VIPPR Routing Table ---\n");
    r=routers_anchor;
    while (r!=NULL) {
	printf("Router %d\n",r->id);
	rr=r->rtr;
	while (rr!=NULL) {
	    printf("\t%16s/",inet_ntoa(*((struct in_addr *)&(rr->dest))));
	    printf("%16s ->\t",inet_ntoa(*((struct in_addr *)&(rr->mask))));
	    printf("%16s\n",inet_ntoa(rr->gwip));
	    rr=rr->next;
	}
	r=r->next;
    }
}

route_t *router_findnexthop(vip_t *vip, struct in_addr *ip) {
    route_t	*r;

    r=vip->rtab;
    while (r!=NULL) {
	if ((*((u_int32_t *)ip) & r->mask) == r->dest) 
	    return r;
	else 
	    r=r->next;
    }
    return NULL;
}


/* *********************
 * INTERFACES
 * *********************/

/* returns the new interface structure or NULL if something, probably 
 * init_device() went wrong. This enables the caller vip_add() to include 
 * the pointer in the new VIP structure */
interface_t *interface_add(char *name) {
    interface_t		*n,*c;
    
    /* if we already have this interface, return it */
    if ((c=interface_findbyname(name))!=NULL) return c;

    /* create a new interface structure */
    n=smalloc(sizeof(interface_t));
    n->name=smalloc(strlen(name)+1);
    strcpy(n->name,name);

    if (init_device(n)!=0) {
	free(n->name);
	free(n);
	return NULL;
    }

    if (if_anchor==NULL) {
	if_anchor=n;
    } else {
	c=if_anchor;
	while (c->next!=NULL) c=c->next;
	c->next=n;
    }

    if ((cfg.verbose>2)||(cfg.debug)) 
	printf("interface_add(): added %s\n",n->name);

    return n;
}

    
/* returns a pointer to the interface structure or NULL if not found */
interface_t *interface_findbyname(char *name) {
    interface_t		*c;

    c=if_anchor;
    while (c!=NULL) {
	if (strcmp(c->name,name)==0) return c;
	c=c->next;
    }

    return NULL;
}

void interface_destroy(void) {
    interface_t		*c;

    c=if_anchor;
    while (c!=NULL) {
	if_anchor=c;
	c=c->next;
	if (cfg.debug) printf("interface_destroy(): removing %s\n",
		if_anchor->name);
	/* close socket */
	close(if_anchor->sfd);
	/* free name */
	if (if_anchor->name!=NULL) free(if_anchor->name);
	/* free structure */
	free(if_anchor);
    }
}

void interface_send(interface_t *intf, u_char *p, unsigned int len) {
    queue_t		*q,*c;

    q=smalloc(sizeof(queue_t));
    q->pack=smalloc(len);
    memcpy(q->pack,p,len);
    q->plen=len;

    if ((c=intf->outq)==NULL) {
	intf->outq=q;
    } else {
	while (c->next!=NULL) c=c->next;
	c->next=q;
    }
}

/* *********************
 * ARP cache
 * *********************/

arpcache_t *arp_findbyip(struct in_addr *ip) {
    arpcache_t	*a;

    a=arp_anchor;
    while (a!=NULL)
	if (!memcmp(&(a->addr),ip,IP_ADDR_LEN)) return a; else a=a->next;
    return NULL;
}

arpcache_t *arp_new(struct in_addr *ip) {
    arpcache_t		*a,*c;

    if ((a=arp_findbyip(ip))!=NULL) return a;

    a=smalloc(sizeof(arpcache_t));
    if (arp_anchor==NULL) {
	arp_anchor=a;
    } else {
	c=arp_anchor;
	while (c->next!=NULL) c=c->next;
	c->next=a;
    }
    memcpy(&(a->addr),ip,IP_ADDR_LEN);
    a->flag=ARP_UNKNOWN;
    a->tx=(unsigned long)time(NULL);

    if (cfg.debug) printf("arp_new(): %s\n",inet_ntoa(*ip));

    return a;
}

/* add a request to all interfaces you can think of ... 
 * the entry in arptable has only the ip addr in it. 
 * Mac and interface are empty */
void arp_request_all(arpcache_t *a) {
    interface_t		*intf;
    vip_t		*v;

    /* for all interfaces, find a VIP that uses this interface and send 
     * out an ARP request with this sender information */
    intf=if_anchor;
    while (intf!=NULL) {
	v=vip_anchor;
	while (v!=NULL) {
	    if ( 
		(*((u_int32_t *)&(a->addr))&*((u_int32_t *)&(v->mask)))==
		(*((u_int32_t *)&(v->addr))&*((u_int32_t *)&(v->mask)))
		) {
	    //if (v->intf==intf) {
		pack_arp_request(v,&(a->addr));
		v=NULL;
	    } else {
		v=v->next;
	    }
	}
	intf=intf->next;
    }

    if (a->flag<ARP_TOREMOVE) a->flag++;
    a->tx=(unsigned long)time(NULL)+TIMER_ARP_REQUESTS;
    if (cfg.debug) printf("arp_request_all(): %s\n",inet_ntoa(a->addr));
}

void arp_remove(arpcache_t *b) {
    arpcache_t	*a;

    if ((a=arp_anchor)==b) {
	arp_anchor=arp_anchor->next;
	if (cfg.debug) 
	    printf("arp_remove(): %s removed (anchor)\n",inet_ntoa(b->addr));
	free(b);
	return;
    }  
    while (a!=NULL) {
	if (a->next==b) {
	    a->next=b->next;
	    if (cfg.debug) 
		printf("arp_remove(): %s removed\n",inet_ntoa(b->addr));
	    free(b);
	    return;
	}
	a=a->next;
    }
}

arpcache_t *arp_set(interface_t *intf, struct in_addr *ip, 
	struct ether_addr *mac) {
    arpcache_t		*a;

    a=arp_anchor;
    while (a!=NULL) {
	if (!memcmp(&(a->addr),ip,IP_ADDR_LEN)) {
	    memcpy(&(a->mac),mac,ETH_ALEN);
	    a->intf=intf;
	    a->flag=ARP_REVIEW;
	    a->tx=(unsigned long)time(NULL)+TIMER_ARP_REVIEW;
	    return a;
	}
	a=a->next;
    } 
    return NULL;
}

void arp_answer_request(interface_t *intf, struct in_addr *sip, 
	struct ether_addr *smac, struct in_addr *dip) {
    vip_t	*v;

    v=vip_anchor;
    while (v!=NULL) {
	if (
		(v->intf==intf) 
		&& (!memcmp(&(v->addr),dip,IP_ADDR_LEN))
		) {
	    if (cfg.debug) printf("arp_answer_request(%s): response to %s\n",
		    intf->name,inet_ntoa(*sip));
	    pack_arp_response(v,smac,sip);
	    return;
	}
	v=v->next;
    }
}

void check_arp_table(void) {
    arpcache_t		*a,*b;

    a=arp_anchor;
    while (a!=NULL) {
	if ( (a->tx<=(unsigned int)time(NULL)) && (a->flag!=ARP_TOREMOVE) ) {
	    if (a->flag==ARP_REVIEW) a->flag=ARP_REQ1;
	    arp_request_all(a);
	}
	a=a->next;
    }

    /* remove all that are flagge ARP_TOREMOVE */
    a=arp_anchor;
    while (a!=NULL) {
	b=a;
	a=a->next;
	if (b->flag==ARP_TOREMOVE) arp_remove(b);
    }
}

void print_arp_table(void) {
    arpcache_t		*a;

    printf("--- ARP Table ---\n");
    a=arp_anchor;
    while (a!=NULL) {
	if (a->intf!=NULL) {
	    printf("%16s %02x:%02x:%02x:%02x:%02x:%02x %d %ld on %s\n",
		    inet_ntoa(a->addr),
		    a->mac.ether_addr_octet[0],
		    a->mac.ether_addr_octet[1],
		    a->mac.ether_addr_octet[2],
		    a->mac.ether_addr_octet[3],
		    a->mac.ether_addr_octet[4],
		    a->mac.ether_addr_octet[5],
		    a->flag,
		    a->tx,
		    a->intf->name);
	} else {
	    printf("%16s [unresolved] %d %ld\n",
		    inet_ntoa(a->addr),
		    a->flag,
		    a->tx);
	}
	a=a->next;
    }
}
