/* VIPPR network functions 
 *
 * $Id: vippr_net.c,v 1.3 2001/07/10 04:24:04 fx Exp $
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

/* "new" packet socket implemented according to packet(7) */
#include <sys/socket.h>
#include <features.h>           /* for the glibc version number */
#if __GLIBC__ >= 2 && __GLIBC_MINOR >= 1
#include <netpacket/packet.h>
#include <net/ethernet.h>       /* the L2 protocols */
#else
#include <asm/types.h>
#include <linux/if_packet.h>
#include <linux/if_ether.h>     /* The L2 protocols */
#endif
#include <sys/types.h>
#include <sys/param.h>
#include <sys/file.h>
#include <sys/ioctl.h>
#include <sys/socket.h>
#include <net/if.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <errno.h>

#include <protocols.h>
#include <vippr.h>

/* returns 0 if ok or -1 if not
 *
 * IP address, MAC and socket descriptor are saved to the structure *intf 
 * where the interface name is already filled in and passed this way to 
 * init_device()
 */
int init_device(interface_t *intf) {
    int			tempsock;
    int			dev_found=0;
    struct ifconf	ifc;
    struct ifreq        ibuf[16], *ifrp, *ifr_last, ifr_temp;
    // variables used to bind(2) the socket
    struct sockaddr_ll  sall;
    int			dev_idx;
    struct packet_mreq  mreq;


    if ((intf==NULL)||(intf->name==NULL)) return (-1);
    if ((tempsock=socket(AF_INET,SOCK_DGRAM,0))<0) {
	perror("socket(tempsock)");
	return (-1);
    }

    ifc.ifc_len=sizeof(ibuf);
    ifc.ifc_buf=(caddr_t)ibuf;
    memset((char *)ibuf,0,sizeof(ibuf));
    if (
            (ioctl(tempsock,SIOCGIFCONF,(char *)&ifc)<0)
            ||
            ((unsigned int)ifc.ifc_len < sizeof(struct ifreq))
            ) {
	perror("ioctl(IFCONF)");
	return (-1);
    }

    ifrp=ibuf;
    ifr_last=(struct ifreq *)((char *)ibuf+ifc.ifc_len);
    dev_idx=(-1);
    do {
	if (strcmp(ifrp->ifr_name,intf->name)==0) {
	    memset(&ifr_temp,0,sizeof(ifr_temp));
	    strncpy(ifr_temp.ifr_name,ifrp->ifr_name,
		    sizeof(ifr_temp.ifr_name));
	    ioctl(tempsock,SIOCGIFINDEX,&ifr_temp);
	    dev_idx=ifr_temp.ifr_ifindex;
	    dev_found=1;
	    if (cfg.verbose>1) printf("using device %s\n",ifrp->ifr_name);
	    break;
	} else {
	    ifrp++;
	}
    } while (ifrp!=ifr_last);

    if (!dev_found) { 
	fprintf(stderr,"Device '%s' not found\n",intf->name); 
	return (-1); 
    }

    /* get HW addr */
    memset(&ifr_temp,0,sizeof(ifr_temp));
    strncpy(ifr_temp.ifr_name,ifrp->ifr_name,
	    sizeof(ifr_temp.ifr_name));
    if (ioctl(tempsock, SIOCGIFHWADDR, &ifr_temp) < 0 ) {
	perror("ioctl(IFHWADDR)");
	return (-1);
    }
    memcpy(&(intf->mac),&ifr_temp.ifr_hwaddr.sa_data,ETH_ALEN);
    /* get IP */
    memset(&ifr_temp,0,sizeof(ifr_temp));
    strncpy(ifr_temp.ifr_name,ifrp->ifr_name,
	    sizeof(ifr_temp.ifr_name));
    if (ioctl(tempsock, SIOCGIFADDR, &ifr_temp) < 0 ) {
	perror("ioctl(IFADDR)");
	return (-1);
    }
    memcpy(&(intf->addr),
	    &(*(struct sockaddr_in *)&ifr_temp.ifr_addr).sin_addr.s_addr,
	    IP_ADDR_LEN);
    close(tempsock);

    if ((intf->sfd=socket(PF_PACKET, SOCK_RAW, htons(ETH_P_ALL)))<0) {
	perror("socket(PF_PACKET)");
	return (-1);
    }
    intf->idx=dev_idx;

    /* send into all-multicast mode */
    memset(&mreq,0,sizeof(mreq));
    mreq.mr_ifindex=dev_idx;
    mreq.mr_type=PACKET_MR_ALLMULTI;
    if (setsockopt(intf->sfd,SOL_PACKET,PACKET_ADD_MEMBERSHIP,
	    &mreq,sizeof(mreq))!=0) {
	perror("setsockopt(PACKET_MR_ALLMULTI)");
    } 

    /* send into promiscuous mode */
    memset(&mreq,0,sizeof(mreq));
    mreq.mr_ifindex=dev_idx;
    mreq.mr_type=PACKET_MR_PROMISC;
    if (setsockopt(intf->sfd,SOL_PACKET,PACKET_ADD_MEMBERSHIP,
	    &mreq,sizeof(mreq))!=0) {
	perror("setsockopt(PACKET_MR_PROMISC)");
    } 

    /* bind it */
    sall.sll_family=AF_PACKET;
    sall.sll_ifindex=dev_idx;
    sall.sll_protocol=htons(ETH_P_ALL);
    sall.sll_pkttype=
        PACKET_HOST
        |PACKET_OTHERHOST
        |PACKET_BROADCAST
        |PACKET_OUTGOING
        |PACKET_MULTICAST;
    
    if (bind(intf->sfd,(struct sockaddr *)&sall,sizeof(struct sockaddr_ll))<0) {
	perror("bind()");
	return (-1);
    }
    makenonblock(intf->sfd);

    if (cfg.verbose) 
	printf("VIPPR on %s [%02x:%02x:%02x:%02x:%02x:%02x] activated\n",
		inet_ntoa(intf->addr),
		intf->mac.ether_addr_octet[0],
		intf->mac.ether_addr_octet[1],
		intf->mac.ether_addr_octet[2],
		intf->mac.ether_addr_octet[3],
		intf->mac.ether_addr_octet[4],
		intf->mac.ether_addr_octet[5]);

    return 0;
}


/* receives a packet from the wire (promiscuous mode), identifies the 
 * Ethernet frame format and the payload.
 *
 * If the packet is IP or ARP, the size parameter is reduced to reflect the 
 * new one and the pointer together with size and type are passed back to the
 * caller. 
 * The original packet is still available in sys.packet for bridging/routing */
u_char *next_packet(int sfd, unsigned int *ptype, int *psize) {
    u_char		*ip;
    
    memset(sys.packet,0,PACKET_MAX);
    if ((*psize=recvfrom(sfd,sys.packet,PACKET_MAX,0,NULL,NULL))<1) 
	return NULL; 
    sys.plen=*psize;
    /* OK, so it's a packet */
    // if (*psize<15) { /* can't have any payload */ return NULL; }
    if (*psize<32) { 
	/* Ethernet dwarf */ 
	if (cfg.debug) printf("next_packet(): eth dwarf, %d (drop)\n",*psize);
	return NULL; 
    }

    if (ntohs(*((u_int16_t *)((void *)sys.packet+12)))>0x05dc) { 
	/* ETHERNET II */
	if ((cfg.debug)&&(cfg.verbose>1)) 
		printf("next_packet(): Frame: ETH II\n");
	ip=(u_char *)((void *)sys.packet+14);
	*ptype=htons(*((u_int16_t *)((void *)sys.packet+12)));
	*psize-=14;
    } else {
	if ((cfg.debug)&&(cfg.verbose>1)) 
	    printf("next_packet(): Frame: fritte (drop)\n");
	ip=NULL;
	*ptype=0;
    }
    return ip;
}


void net_ifrecv(interface_t *intf) {
    unsigned int	psize=0;
    int			ptype=0;
    u_char		*pp;
    iphdr_t		*ipp;
    vip_t		*v_mac;
    arphdr_t		*arpp;

    /* STEP 1 - no frame - no processing */
    if ((pp=next_packet(intf->sfd,&ptype,&psize))==NULL) return; /* OUT */

    /* STEP 2
     * if this frame is addressed to our real IF, return here because this 
     * saves us from going through the whole list of VIPs */

    if (!cfg.disable_fastdrop) {
	if (
		(!memcmp(sys.packet,&(intf->mac),ETH_ALEN))
		||(!memcmp((void *)sys.packet+6,&(intf->mac),ETH_ALEN))
		) {
	    if ((cfg.debug) && (cfg.verbose>1))
		printf("net_ifrecv(%s): fastdrop (drop)\n",intf->name);
	    return; // OUT 
	} 
    }

    /* STEP 3
     * if this frame is a broadcast, let's have a look at it 
     */
    if (!memcmp(sys.packet,bcastaddr,ETH_ALEN)) {
	if (cfg.debug) printf("net_ifrecv(%s): bcast found\n",intf->name);
	/* as Broadcast, we only look at ARP requests, so ... */
	if (ptype!=ETHERTYPE_ARP) {
	    if (cfg.debug) 
		printf("net_ifrecv(%s): bcast not ARP (drop)\n",intf->name);
	    return; /* OUT */
	}
	/* +++++++++++ handle ARP ++++++++++++ */
	arpp=(arphdr_t *)pp;
	if (ntohs(arpp->opcode)==ARPOP_REQUEST) {
	    arp_answer_request(intf,
		    (struct in_addr *)&(arpp->sip),
		    (struct ether_addr *)&(arpp->sha),
		    (struct in_addr *)&(arpp->tip));
	} 
	/* else { printf("ARP != request to bcast ????\n"); } */
    }

    /* STEP 4
     * if the frame is not addressed to one of or VIP's MACs, 
     * we can drop it
     */
    if ((v_mac=vip_findbymacif((struct ether_addr *)sys.packet,intf))==NULL) {
	if ((cfg.debug) && (cfg.verbose>1))
	    printf("net_ifrecv(%s): not to VIP MAC"
		    " [%02x:%02x:%02x:%02x:%02x:%02x] (drop)\n",intf->name,
		    *sys.packet&0xFF, 
		    *(sys.packet+1)&0xFF,
		    *(sys.packet+2)&0xFF,
		    *(sys.packet+3)&0xFF,
		    *(sys.packet+4)&0xFF,
		    *(sys.packet+5)&0xFF);
	return; /* OUT */
    } else if (cfg.debug)
	    printf("net_ifrecv(%s): packet addressed to VIP MAC %s\n",
		    intf->name,inet_ntoa(v_mac->addr));
    
    /* STEP 5
     * if this is IP, find out whether this is actually addressed to 
     * our VIPs IP-address */
    if (ptype==ETHERTYPE_IP) {
	ipp=(iphdr_t *)pp;
	/* if this packet has our interface as final destination in 
	 * layer 2 and 3, handle it somehow */
	if (!memcmp(&(v_mac->addr),&(ipp->daddr),IP_ADDR_LEN)) {
		handle_incomming_vip(pp,psize,v_mac);
		return; /* OUT */
	}

/* ----- ROUTING STEP HERE ------ */
	arp_new(&(ipp->saddr));
	arp_set(intf,&(ipp->saddr),(struct ether_addr *)((void *)sys.packet+6));
	/* MAC addr points to a VIP, IP address is not this VIP
	 * 	ERGO:   R O U T E  I T  !!  */
	if (v_mac->viptype==VIP_TYPE_GRE) 
	    net_route_gre(v_mac);
	else 
	    net_route_packet(v_mac);

    } else if (ptype==ETHERTYPE_ARP) {
	arpp=(arphdr_t *)pp;
	if (ntohs(arpp->opcode)==ARPOP_REQUEST) {
	    /* handle ar prequest for our VIP MAC */
	    arp_answer_request(intf,
		    (struct in_addr *)&(arpp->sip),
		    (struct ether_addr *)&(arpp->sha),
		    (struct in_addr *)&(arpp->tip));
	} else if (ntohs(arpp->opcode)==ARPOP_REPLY) {
	    /* add response to ARP cache */
	    if (arp_set(intf,
			(struct in_addr *)&(arpp->sip),
			(struct ether_addr *)&(arpp->sha))==NULL) 
		/* hacker hacker hacker */
		fprintf(stderr,"WARNING: ARP response to IP not asked for:"
			"%s [%02x:%02x:%02x:%02x:%02x:%02x]\n",
			inet_ntoa(*((struct in_addr *)&(arpp->sip))),
			arpp->sha[0], arpp->sha[1], arpp->sha[2],
			arpp->sha[3], arpp->sha[4], arpp->sha[5]);
	    else 
		if (cfg.debug)
		    printf("net_ifrecv(%s): ARP response received for %s\n",
			    intf->name,
			    inet_ntoa(*((struct in_addr *)&(arpp->sip))));
	} else {
	    fprintf(stderr,"net_ifrecv(%s): unknown ARP op code\n",intf->name);
	}
    } else if (ptype==ETHERTYPE_REVARP) {
	fprintf(stderr,"net_ifrecv(%s): REVARP not supported\n",intf->name);
    } else {
	fprintf(stderr,"net_ifrecv(%s): Ethernet Protocol %d ?!?\n",
		intf->name,ptype);
    }
}

/* sends the first frame from the interface's outq onto the wire */
void net_ifsend(interface_t *intf) {
    struct sockaddr_ll 	sall;
    int			slen;
    queue_t		*q;
    unsigned int	ss;

    /* if there is nothing to send, return ;) */
    if (intf->outq==NULL) return;
    q=intf->outq->next;

    slen=sizeof(sall);
    sall.sll_family = AF_PACKET;
    sall.sll_ifindex = intf->idx;
    if ((ss=sendto(intf->sfd,intf->outq->pack,intf->outq->plen,0,
	    (struct sockaddr*) &sall,slen))<0) {
	perror("sendto()");
    } else if (cfg.debug) 
	printf("net_ifsend(%s): %d bytes to wire\n",intf->name,ss);

    /* free this memory and reattach whatever was on outq->next before */
    free(intf->outq->pack);
    free(intf->outq);
    intf->outq=q;
}


/* this is the function that will handle "services" later */
void    handle_incomming_vip(u_char *packet,unsigned int psize,vip_t *vip) {
    /* I'M a router, not a fucking windows box */
    if ((cfg.verbose>1)||(cfg.debug)) { 
	iphdr_t	*ip=(iphdr_t *)packet;
	printf("handle_incomming_vip(): incomming packet from %s to ",
		inet_ntoa(ip->saddr));
	printf("%s\n",inet_ntoa(ip->daddr));
    }
    return;
}


/* Function does:
 * 	- find a next hop for the packet in this VIP router 
 * 		- if not, send an ICMP net unreach back
 * 	- find the next hop MAC
 * 		- if not, drops packet
 * 	- change the dest MAC and send
 */
void net_route_packet(vip_t *v) {
    iphdr_t	*ip;
    route_t	*r,*r2;
    vip_t	*outi;
    arpcache_t	*arp;

    ip=(iphdr_t *)((void *)sys.packet+sizeof(etherIIhdr_t));
    if (v->rtab==NULL) {
	if (cfg.debug) {
	    printf("net_route_packet(): No router to %s",
		    inet_ntoa(ip->daddr));
	    printf(" on VIP %s\n",
		    inet_ntoa(v->addr));
	}
	return; /* OUT */
    }

    if ((r=router_findnexthop(v,&(ip->daddr)))==NULL) {
	if (cfg.debug) 
	    printf("net_route_packet(): No route to host %s\n",
		    inet_ntoa(ip->daddr));

	/* send icmp unreach */
	return;
    }  

    if (cfg.debug) {
	printf("net_route_packet(): next hop %s", inet_ntoa(r->gwip));
	printf(" for dest %s\n", inet_ntoa(ip->daddr));
    }

    /* may be the next hop addr is a VIP of us */
    if ((outi=vip_findbyip(&(r->gwip)))!=NULL) {
	if (cfg.debug) 
	    printf("net_route_packet(): found the VIP %s =>> %s\n",
		inet_ntoa(outi->addr),outi->intf->name);
	/* MAC known ?*/
	if ((arp=arp_findbyip(&(ip->daddr)))!=NULL)  {
	    /* copy destination in orig packet */
	    memcpy((u_char*)sys.packet,&(arp->mac),ETH_ALEN);
	    /* copy our src MAC in orig packet */
	    memcpy((u_char*)(sys.packet+ETH_ALEN),&(v->mac),ETH_ALEN);
	    if ((cfg.debug)&&(cfg.verbose)) lamont_hdump(sys.packet,sys.plen);
	    interface_send(outi->intf,sys.packet,sys.plen);
	    if (cfg.verbose>1) { printf("+"); fflush(stdout); }

	} else {
	    if (cfg.debug) 
		printf("net_route_packet(): sorry, have to ARP for %s\n",
			inet_ntoa(ip->daddr));
	    arp_new(&(ip->daddr));
	}
    } else {
	/* send to next best gateway */
	if (cfg.debug) printf("net_route_packet(): Transit packet\n");
	if ((arp=arp_findbyip(&(r->gwip)))!=NULL)  {
	    /* copy dest MAC in orig packet */
	    memcpy((u_char*)sys.packet,&(arp->mac),ETH_ALEN);
	    /* copy our src MAC in orig packet */
	    memcpy((u_char*)(sys.packet+ETH_ALEN),&(v->mac),ETH_ALEN);
	    if ((cfg.debug)&&(cfg.verbose)) lamont_hdump(sys.packet,sys.plen);
	    if (
		    ((r2=router_findnexthop(v,&(r->gwip)))==NULL)
		    || 
		    ((outi=vip_findbyip(&(r2->gwip)))==NULL)
		    ) {
		fprintf(stderr,"Could not determine destination VIP IF for"
			" gateway %s\n",inet_ntoa(r->gwip));
	    } else {
		interface_send(outi->intf,sys.packet,sys.plen);
		if (cfg.verbose>1) { printf("*"); fflush(stdout); }
	    }
	} else {
	    if (cfg.debug) 
		printf("net_route_packet(): sorry, have to ARP for GW %s\n",
			inet_ntoa(r->gwip));
	    arp_new(&(r->gwip));
	}
    }
}

void net_route_gre(vip_t *v) {
    iphdr_t		*ip,*ipgre;
    route_t		*r,*r2;
    vip_t		*outi;
    arpcache_t		*arp;
    u_char		*np;
    grehdr_t		*gre;
    etherIIhdr_t	*ee;
    unsigned int 	nplen;
    void		*fill;

    // ip=(iphdr_t *)((void *)sys.packet+sizeof(etherIIhdr_t));

    if (v->ex.gre.flags==VIP_GRE_TUNNELKEY) {
	nplen=sys.plen+sizeof(iphdr_t)+sizeof(grehdr_t)+sizeof(u_int32_t);
    } else {
	nplen=sys.plen+sizeof(iphdr_t)+sizeof(grehdr_t);
    }
    np=smalloc(nplen);
    ee=(etherIIhdr_t *)np;
    memcpy(np,sys.packet,sizeof(etherIIhdr_t));

    ipgre=(iphdr_t *)((void*)np+sizeof(etherIIhdr_t));
    ipgre->version=4;
    ipgre->ihl=sizeof(iphdr_t)/4;
    ipgre->protocol=IPPROTO_GRE;
    ipgre->id=htons((unsigned int)(65535.0*rand()/(RAND_MAX+1.0)));
    ipgre->ttl=0x40;
    memcpy(&(ipgre->saddr.s_addr),
	    &(v->ex.gre.tunnelsource.s_addr),IP_ADDR_LEN);
    memcpy(&(ipgre->daddr.s_addr),
	    &(v->ex.gre.tunneldestination.s_addr),IP_ADDR_LEN);
    gre=(grehdr_t *)((void*)np+sizeof(iphdr_t)+sizeof(etherIIhdr_t));

    if (v->ex.gre.flags==VIP_GRE_TUNNELKEY) {
	ipgre->tot_len=htons(sys.plen-sizeof(etherIIhdr_t)+
		sizeof(iphdr_t)+sizeof(grehdr_t)+sizeof(u_int32_t));
	gre->flags=htons(0x2000); /* key present */
	gre->proto=htons(0x0800);
	*((u_int32_t *)((void*)np+sizeof(iphdr_t)
		    +sizeof(etherIIhdr_t)+sizeof(grehdr_t)))=
	    htonl(v->ex.gre.tunnelkey);
	fill=(void*)np+sizeof(etherIIhdr_t)+
		sizeof(iphdr_t)+sizeof(grehdr_t)+sizeof(u_int32_t);
	memcpy(fill,(void *)sys.packet+sizeof(etherIIhdr_t),
		sys.plen-sizeof(etherIIhdr_t));
    } else {
	ipgre->tot_len=htons(sys.plen-sizeof(etherIIhdr_t)
		+sizeof(iphdr_t)+sizeof(grehdr_t));
	gre->flags=0;
	gre->proto=htons(0x0800);
	fill=(void*)np+sizeof(etherIIhdr_t)+sizeof(iphdr_t)+sizeof(grehdr_t);
	memcpy(fill,(void *)sys.packet+sizeof(etherIIhdr_t),
		sys.plen-sizeof(etherIIhdr_t));
    }
    ipgre->check=chksum((u_char *)(ipgre),sizeof(iphdr_t));
    /* ARHLSIHSDIH! Don't route the original ! */
    ip=ipgre;

    if (v->rtab==NULL) {
	if (cfg.debug) {
	    printf("net_route_gre(): No router to %s",
		    inet_ntoa(ip->daddr));
	    printf(" on VIP %s\n",
		    inet_ntoa(v->addr));
	}
	free(np);
	return; /* OUT */
    }

    if ((r=router_findnexthop(v,&(ip->daddr)))==NULL) {
	if (cfg.debug) 
	    printf("net_route_gre(): No route to host %s\n",
		    inet_ntoa(ip->daddr));

	/* send icmp unreach */
	free(np);
	return;
    } 

    if (cfg.debug) {
	printf("net_route_gre(): next hop %s", inet_ntoa(r->gwip));
	printf(" for dest %s\n", inet_ntoa(ip->daddr));
    }

    /* may be the next hop addr is a VIP of us */
    if ((outi=vip_findbyip(&(r->gwip)))!=NULL) {
	if (cfg.debug) 
	    printf("net_route_gre(): found the VIP %s =>> %s\n",
		inet_ntoa(outi->addr),outi->intf->name);
	/* MAC known ?*/
	if ((arp=arp_findbyip(&(ip->daddr)))!=NULL)  {
	    memcpy((u_char*)np,&(arp->mac),ETH_ALEN);
	    /* copy our src MAC in orig packet */
	    memcpy((u_char*)np+ETH_ALEN,&(v->mac),ETH_ALEN);
	    if ((cfg.debug)&&(cfg.verbose)) lamont_hdump(np,nplen);
	    interface_send(outi->intf,np,nplen);
	    if (cfg.verbose>1) { printf("T(+)"); fflush(stdout); }
	} else {
	    if (cfg.debug) 
		printf("net_route_gre(): sorry, have to ARP for %s\n",
			inet_ntoa(ip->daddr));
	    arp_new(&(ip->daddr));
	}
    } else {
	/* send to next best gateway */
	if (cfg.debug) printf("net_route_gre(): Transit packet\n");
	if ((arp=arp_findbyip(&(r->gwip)))!=NULL)  {
	    memcpy((u_char*)np,&(arp->mac),ETH_ALEN);
	    /* copy our src MAC in orig packet */
	    memcpy((u_char*)np+ETH_ALEN,&(v->mac),ETH_ALEN);
	    if ((cfg.debug)&&(cfg.verbose)) lamont_hdump(np,nplen);
	    if (
		    ((r2=router_findnexthop(v,&(r->gwip)))==NULL)
		    || 
		    ((outi=vip_findbyip(&(r2->gwip)))==NULL)
		    ) {
		fprintf(stderr,"Could not determine destination VIP IF for"
			" gateway %s\n",inet_ntoa(r->gwip));
	    } else {
		interface_send(outi->intf,np,nplen);
		if (cfg.verbose>1) { printf("T(*)"); fflush(stdout); }
	    }
	} else {
	    if (cfg.debug) 
		printf("net_route_gre(): sorry, have to ARP for GW %s\n",
			inet_ntoa(r->gwip));
	    arp_new(&(r->gwip));
	}
    }
    free(np);
}

void    makenonblock(int s) {
    int         rc=1;

    rc= O_NONBLOCK | fcntl(s, F_GETFL);
    fcntl(s,F_SETFL,rc);
}

