/* VIPPR packet functions 
 *
 * $Id: vippr_packs.c,v 1.2 2001/06/12 14:25:51 fx Exp $
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/param.h>
#include <sys/file.h>
#include <sys/ioctl.h>
#include <sys/socket.h>
#include <net/if.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <errno.h>

#include <protocols.h>
#include <vippr.h>

void pack_arp_request(vip_t *sender,struct in_addr *who) {
    u_char	*pack;
    arphdr_t	*a;

    pack=smalloc(sizeof(etherIIhdr_t)+sizeof(arphdr_t));
    pack_mketh(pack,sender,(struct ether_addr *)bcastaddr,ETHERTYPE_ARP);
    a=(arphdr_t *)((void *)pack+sizeof(etherIIhdr_t));
    a->hardware=htons(1);
    a->protocol=htons(0x0800);
    a->hw_size=ETH_ALEN;
    a->proto_size=IP_ADDR_LEN;
    a->opcode=htons(ARPOP_REQUEST);
    memcpy(&(a->sha),&(sender->mac),ETH_ALEN);
    memcpy(&(a->sip),&(sender->addr.s_addr),IP_ADDR_LEN);
    /* Only SUN does this:  memcpy(&(a->tha),bcastaddr,ETH_ALEN); */
    memcpy(&(a->tip),&(who->s_addr),IP_ADDR_LEN);

    /* send (read: put into queue) */
    interface_send(sender->intf,pack,sizeof(etherIIhdr_t)+sizeof(arphdr_t));
    free(pack);
} 

void pack_arp_response(vip_t *sender,struct ether_addr *dmac,
	struct in_addr *dip) {
    u_char	*pack;
    arphdr_t	*a;
#define ARPREPLYS	(sizeof(etherIIhdr_t)+sizeof(arphdr_t)+18)

    pack=smalloc(ARPREPLYS);
    pack_mketh(pack,sender,dmac,ETHERTYPE_ARP);
    a=(arphdr_t *)((void *)pack+sizeof(etherIIhdr_t));
    a->hardware=htons(1);
    a->protocol=htons(0x0800);
    a->hw_size=ETH_ALEN;
    a->proto_size=IP_ADDR_LEN;
    a->opcode=htons(ARPOP_REPLY);
    memcpy(&(a->sha),&(sender->mac),ETH_ALEN);
    memcpy(&(a->sip),&(sender->addr.s_addr),IP_ADDR_LEN);
    memcpy(&(a->tha),dmac,ETH_ALEN);
    memcpy(&(a->tip),&(dip->s_addr),IP_ADDR_LEN);

    /* send (read: put into queue) */
    interface_send(sender->intf,pack,ARPREPLYS);
    free(pack);
} 


/* makes the Ethernet II header into the memory area pointed at with p.
 * WARNING: Does not check mem bounds !!!!!
 * This could be implemented as macro but typechecks help to debug crapy 
 * brains */
void pack_mketh(u_char *p, vip_t *sender, struct ether_addr *dest,
	unsigned int ptype) {
    memcpy(p,dest,ETH_ALEN);
    memcpy(p+6,&(sender->mac),ETH_ALEN);
    *((u_int16_t*)(p+12))=htons(ptype);
}
    
/* returns the checksum 
 * WARNING: if left over bytes are present, the memory after *data has to
 * contain 0x00 series and should be part of the buffer
 * -> make the buffer for data at least count+1 bytes long ! */
u_int16_t chksum(u_char *data, unsigned long count) {
    u_int32_t           sum = 0;
    u_int16_t           *wrd;

    wrd=(u_int16_t *)data;
    while( count > 1 )  {
        sum = sum + *wrd;
        wrd++;
        count -= 2;
    }

    /*  Add left-over byte, if any */
    if( count > 0 ) 
        sum = sum + ((*wrd &0xFF)<<8);

    /*  Fold 32-bit sum to 16 bits */
    while (sum>>16) 
        sum = (sum & 0xffff) + (sum >> 16);

    return (~sum);
}


